package paths

case class Coord(x: Int, y: Int)

def countPathsNaive(
    src: Coord,
    dst: Coord,
    hasObstacle: Coord => Boolean
): Int =
  require(!hasObstacle(src) && !hasObstacle(dst))
  def countFrom(src: Coord): Int =
    if src == dst then 1
    else if hasObstacle(src) || src.x > dst.x || src.y > dst.y then 0
    else
      math.addExact(
        countFrom(src.copy(x = src.x + 1)),
        countFrom(src.copy(y = src.y + 1))
      )
  countFrom(src)

def countPathsFastRec(
    src: Coord,
    dst: Coord,
    hasObstacle: Coord => Boolean
): Int =
  require(!hasObstacle(src) && !hasObstacle(dst))
  val cache = scala.collection.mutable.Map.empty[Coord, Int]
  def countFrom(src: Coord): Int =
    cache.getOrElseUpdate(
      src,
      if src == dst then 1
      else if hasObstacle(src) || src.x > dst.x || src.y > dst.y then 0
      else
        math.addExact(
          countFrom(src.copy(x = src.x + 1)),
          countFrom(src.copy(y = src.y + 1))
        )
    )
  countFrom(src)

def countPathsFastLoop(
    src: Coord,
    dst: Coord,
    hasObstacle: Coord => Boolean
): Int =
  require(!hasObstacle(src) && !hasObstacle(dst))
  val cache = scala.collection.mutable.Map.empty[Coord, Int].withDefaultValue(0)
  cache(dst) = 1
  for x <- dst.x to src.x by -1 do
    for y <- dst.y to src.y by -1 do
      val c = Coord(x, y)
      if c != dst then
        cache(c) =
          if hasObstacle(c) then 0
          else cache(Coord(x + 1, y)) + cache(Coord(x, y + 1))
  cache(src)

def countPathsFastTopo(
    src: Coord,
    dst: Coord,
    hasObstacle: Coord => Boolean
): Int =
  require(!hasObstacle(src) && !hasObstacle(dst))

  def neighbors(c: Coord) =
    List(c.copy(x = c.x + 1), c.copy(y = c.y + 1)).filter: c =>
      c.x <= dst.x && c.y <= dst.y && !hasObstacle(c)
  def dependents(c: Coord) =
    List(c.copy(x = c.x - 1), c.copy(y = c.y - 1)).filter: c =>
      c.x >= src.x && c.y >= src.y && !hasObstacle(c)

  def bfs(
      boundary: List[Coord],
      previous: List[List[Coord]]
  ): List[List[Coord]] =
    if boundary.contains(src) || boundary.isEmpty then
      previous.reverse
    else
      val next = boundary.flatMap(dependents).distinct
      bfs(next, next :: previous)

  val cache =
    bfs(List(dst), Nil).foldLeft(Map(dst -> 1)): (cache, boundary) =>
      boundary.map(c =>
        c -> neighbors(c).map(n => cache.getOrElse(n, 0)).sum
      ).toMap

  cache.getOrElse(src, 0)

val countPathsFast = countPathsFastTopo
