package paths

import scala.concurrent.duration.Duration

class NaiveTests extends munit.FunSuite:
  test("countPathsNaive: Invalid inputs (1pt)"):
    intercept[IllegalArgumentException]:
      countPathsNaive(Coord(0, 0), Coord(1, 1), _ => true)

  test("countPathsNaive: Already there (1pt)"):
    assertEquals(countPathsNaive(Coord(0, 0), Coord(0, 0), _ => false), 1)
    assertEquals(countPathsNaive(Coord(-2, -2), Coord(-2, -2), _ => false), 1)
    assertEquals(
      countPathsNaive(Coord(100000, 100000), Coord(100000, 100000), _ => false),
      1
    )

  test("countPathsNaive: Never down or to the left (1pt)"):
    assertEquals(countPathsNaive(Coord(10, 10), Coord(10, 9), _ => false), 0)
    assertEquals(countPathsNaive(Coord(10, 10), Coord(9, 10), _ => false), 0)
    assertEquals(countPathsNaive(Coord(100000, 100000), Coord(10, 9), _ => false), 0)
    assertEquals(countPathsNaive(Coord(100000, 100000), Coord(9, 10), _ => false), 0)

  test("countPathsNaive: Immediate neighbors (1pt)"):
    assertEquals(countPathsNaive(Coord(9, 10), Coord(9, 11), _ => false), 1)
    assertEquals(countPathsNaive(Coord(9, 10), Coord(10, 10), _ => false), 1)
    assertEquals(countPathsNaive(Coord(-2, -200000), Coord(-2, -199999), _ => false), 1)
    assertEquals(countPathsNaive(Coord(-2, -200000), Coord(-1, -200000), _ => false), 1)

  test("countPathsNaive: Points on the same line (1pt)"):
    assertEquals(countPathsNaive(Coord(0, 0), Coord(0, 10), _ => false), 1)
    assertEquals(countPathsNaive(Coord(0, 0), Coord(10, 0), _ => false), 1)
    assertEquals(countPathsNaive(Coord(-2, -2), Coord(-2, 5), _ => false), 1)
    assertEquals(countPathsNaive(Coord(-123456, -12), Coord(-123453, -12), _ => false), 1)

  test("countPathsNaive: Arbitrary points close to each other (2pts)"):
    assertEquals(countPathsNaive(Coord(1, 2), Coord(3, 4), _ => false), 6)
    assertEquals(countPathsNaive(Coord(2, 5), Coord(5, 8), _ => false), 20)
    assertEquals(countPathsNaive(Coord(-2, -3), Coord(0, 3), _ => false), 28)
    assertEquals(countPathsNaive(Coord(-123456, -11), Coord(-123450, -7), _ => false), 210)

  test("countPathsNaive: Points on the same line, with obstacles (1pt)"):
    assertEquals(countPathsNaive(Coord(0, 0), Coord(0, 10), _ == Coord(1, 5)), 1)
    assertEquals(countPathsNaive(Coord(0, 0), Coord(10, 0), _ == Coord(4, 1)), 1)
    assertEquals(countPathsNaive(Coord(-2, -2), Coord(-2, 5), _ == Coord(0, 1)), 1)
    assertEquals(countPathsNaive(Coord(-2, -2), Coord(5, -2), _ == Coord(1, 3)), 1)
    assertEquals(countPathsNaive(Coord(0, 0), Coord(0, 10), _ == Coord(0, 5)), 0)
    assertEquals(countPathsNaive(Coord(0, 0), Coord(10, 0), _ == Coord(4, 0)), 0)
    assertEquals(countPathsNaive(Coord(-2, -2), Coord(-2, 5), _ == Coord(-2, 0)), 0)
    assertEquals(countPathsNaive(Coord(-2, -2), Coord(5, -2), _ == Coord(0, -2)), 0)

  test("countPathsNaive: Arbitrary points close to each other, with obstacles (1pt)"):
    val hasObstacle = Set(
      Coord(2, 2),
      Coord(2, 3),
      Coord(3, 6),
      Coord(-1, -1),
      Coord(0, 0),
      Coord(-123456, -10),
      Coord(-123455, -10),
      Coord(-123453, -11)
    ).contains
    assertEquals(countPathsNaive(Coord(1, 2), Coord(3, 4), hasObstacle), 1)
    assertEquals(countPathsNaive(Coord(2, 5), Coord(5, 8), hasObstacle), 8)
    assertEquals(countPathsNaive(Coord(-2, -3), Coord(0, 3), hasObstacle), 9)
    assertEquals(countPathsNaive(Coord(-123456, -11), Coord(-123450, -7), hasObstacle), 35)

  test("countPathsNaive: Arbitrary points close to each other, blocked off (1pt)"):
    def hasObstacle(c: Coord) =
      (c.x + c.y) % 2 == 0
    assertEquals(countPathsNaive(Coord(1, 2), Coord(3, 4), hasObstacle), 0)
    assertEquals(countPathsNaive(Coord(2, 5), Coord(5, 8), hasObstacle), 0)
    assertEquals(countPathsNaive(Coord(-2, -3), Coord(0, 3), hasObstacle), 0)
    assertEquals(countPathsNaive(Coord(-123456, -11), Coord(-123450, -7), hasObstacle), 0)

// Using `Future` in the tests below allows munit to exit after one second

import scala.concurrent.Future
import scala.concurrent.ExecutionContext.Implicits.global

class FastTests extends munit.FunSuite:
  override def munitTimeout: Duration = Duration(1, "second")

  test("countPathsFast: Arbitrary points somewhat far from each other (6pts)"):
    Future:
      assertEquals(countPathsFast(Coord(1, 2), Coord(14, 11), _ => false), 497420)
      assertEquals(countPathsFast(Coord(-2, -2), Coord(12, 6), _ => false), 319770)
      assertEquals(countPathsFast(Coord(2, 5), Coord(24, 15), _ => false), 64512240)
      assertEquals(
        countPathsFast(Coord(-123456, -12), Coord(-123436, -1), _ => false),
        84672315
      )

  test("countPathsFast: Arbitrary points far from each other (6pts)"):
    Future:
      assertEquals(countPathsFast(Coord(1, 2), Coord(14, 16), _ => false), 20058300)
      assertEquals(countPathsFast(Coord(-2, -2), Coord(28, 8), _ => false), 847660528)
      assertEquals(countPathsFast(Coord(2, 5), Coord(26, 15), _ => false), 131128140)
      assertEquals(
        countPathsFast(Coord(-123456, -12), Coord(-123434, 0), _ => false),
        548354040
      )

  val hasObstacle = Set(
    Coord(2, 2),
    Coord(2, 3),
    Coord(3, 6),
    Coord(-1, -1),
    Coord(0, 0),
    Coord(21, 22),
    Coord(25, 36),
    Coord(30, 61),
    Coord(-10, -11),
    Coord(10, 12),
    Coord(-123456, -10),
    Coord(-123455, -10),
    Coord(-123454, -11)
  ).contains

  test(
    "countPathsFast: Arbitrary points somewhat far from each other, with obstacles (6pts)"
  ):
    Future:
      assertEquals(countPathsFast(Coord(1, 2), Coord(14, 11), hasObstacle), 51312)
      assertEquals(countPathsFast(Coord(-2, -2), Coord(12, 6), hasObstacle), 104139)
      assertEquals(countPathsFast(Coord(2, 5), Coord(24, 15), hasObstacle), 33855900)
      assertEquals(
        countPathsFast(Coord(-123456, -12), Coord(-123436, -1), hasObstacle),
        21474180
      )

  test("countPathsFast: Arbitrary points far from each other, with obstacles (6pts)"):
    Future:
      assertEquals(countPathsFast(Coord(1, 2), Coord(14, 16), hasObstacle), 2103024)
      assertEquals(countPathsFast(Coord(-2, -2), Coord(28, 8), hasObstacle), 441642467)
      assertEquals(countPathsFast(Coord(2, 5), Coord(26, 15), hasObstacle), 72120633)
      assertEquals(
        countPathsFast(Coord(-123456, -12), Coord(-123434, 0), hasObstacle),
        141120525
      )

  test("countPathsFast: Arbitrary points very far from each other, blocked off (4pts)"):
    def hasObstacle(c: Coord) = c.x % 5 == 0 || c.y % 5 == 0
    Future:
      assertEquals(countPathsFast(Coord(1, 1), Coord(312, 442), hasObstacle), 0)
      assertEquals(countPathsFast(Coord(2, 4), Coord(231, 527), hasObstacle), 0)
      assertEquals(countPathsFast(Coord(-1, -3), Coord(121, 719), hasObstacle), 0)
      assertEquals(countPathsFast(Coord(-123456, -12), Coord(-123012, 121), hasObstacle), 0)

class StackOverflowTest extends munit.FunSuite:
  override def munitTimeout: Duration = Duration(5, "second")

  test("No stack overflows (harder) (1pt)"):
    Future:
      assertEquals(countPathsFast(Coord(1, 2), Coord(46978, 2), _ => false), 1)
      assertEquals(countPathsFast(Coord(-2, -2), Coord(18413, 0), _ => false), 169583736)

  test("No stack overflows (even harder) (1pt)"):
    Future:
      assertEquals(countPathsFast(Coord(1, 2), Coord(46978, 2), _ => false), 1)
      assertEquals(countPathsFast(Coord(-2, -2), Coord(18413, 0), _ => false), 169583736)
      assertEquals(
        countPathsFast(Coord(0, 0), Coord(12472, 12472), c => c.y < c.x || c.y - c.x > 1),
        1
      )
